#ifdef WIN32
    #include <windows.h>
#else
    #include <sys/time.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>

static float t;

////////////////////////////////////////////////////////////////////////////////
// Utility Code (PRNG, fake functions, math tool functions)
////////////////////////////////////////////////////////////////////////////////

static float dummy = 0.0f;
static void fake2(float x, float y) { dummy += x - y; }
static void fake3(float a, float b, float c) { dummy += a + b - c; }

static unsigned long randSeed1, randSeed2, randSeed3;

static unsigned long randNext(void) {
  randSeed1 =(((randSeed1 & 4294967294UL) << 12) ^ (((randSeed1 << 13) ^ randSeed1) >> 19));
  randSeed2 =(((randSeed2 & 4294967288UL) <<  4) ^ (((randSeed2 <<  2) ^ randSeed2) >> 25));
  randSeed3 =(((randSeed3 & 4294967280UL) << 17) ^ (((randSeed3 <<  3) ^ randSeed3) >> 11));
  return randSeed1 ^ randSeed2 ^ randSeed3;
}

static void randSeed(unsigned long seed) {
  randSeed1 = seed;
  randSeed2 = 0;
  randSeed3 = ~seed;
  randNext();
}

#define randGet()         ((int)(randNext()&0x7FFFFFFF))
#define randFloat()       ((float)(randNext()*(1.0f/4294967296ULL)))
#define randFloatAroundZero(n) (2.0f*(n)*randFloat()-(n))
#define randRange(n)      ((int)(randNext()%(n)))
#define randAroundZero(n) ((int)(randRange((((int)(n))<<1)+1)-(n)))

#define isqrtf(x) (1.0f / sqrtf(x))

////////////////////////////////////////////////////////////////////////////////
// Vortex 2 Cube Grid
////////////////////////////////////////////////////////////////////////////////

#define GRID_SIZE 48
#define CUBE_SIZE 0.475f
#define MOVE_SPEED 5.0f
#define SWITCH_INTERVAL 5.0f

const static unsigned char logo_drg[32] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x82, 0xFE, 0x82, 0x82, 0x7C, 0x00, 0xFE, 0x90, 0x98, 0x94, 0x62, 0x00, 0x7C, 0x82, 0x82, 0x92, 0x5E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
const static unsigned char logo_fm[32] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0x90, 0x90, 0x80, 0x00, 0xFA, 0x00, 0xFE, 0x40, 0x20, 0x40, 0xFE, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
const static unsigned char logo_still[32] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x64, 0x92, 0x92, 0x92, 0x4C, 0x00, 0x80, 0x80, 0xFE, 0x80, 0x80, 0x00, 0xFE, 0x00, 0xFE, 0x02, 0x02, 0x02, 0x00, 0xFE, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00 };
const static unsigned char logo_mds[32] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0x40, 0x20, 0x40, 0xFE, 0x00, 0x82, 0xFE, 0x82, 0x82, 0x7C, 0x00, 0x64, 0x92, 0x92, 0x92, 0x4C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

typedef float (*grid_func)(float, float, float);
static float mat_follow[16], mat_random[16];
#define gf_test gf_ball

static __inline void xform(float *_x, float *_y, float *_z, const float *m) {
    float x = *_x, y = *_y, z = *_z;
    *_x = m[ 0] * x + m[ 4] * y + m[ 8] * z + m[12];
    *_y = m[ 1] * x + m[ 5] * y + m[ 9] * z + m[13];
    *_z = m[ 2] * x + m[ 6] * y + m[10] * z + m[14];
}

static __forceinline float xsample(const unsigned char sample, int y, float f) {
    float s = ((sample >> y) & 1) * (1.0f - f) + ((sample >> (y + 1)) & 1) * f;
    return s*s;
}

static float getlogo(float x, float y, float z, const unsigned char *logo) {
    int ix, iy;
    float fx, fy;
    x += 16.0f;
    y += 4.0f;
    fx = floorf(x);  ix = (int) fx;
    fy = floorf(y);  iy = (int) fy;
    if ((ix < 0) || (iy < 0) || (x >= 32) || (y > 8)) return -1.0f;
    fx = x - fx;
    fy = y - fy;
    fx = xsample(logo[ix], iy, fy) * (1.0f - fx) + xsample(logo[ix + 1], iy, fy) * fx;
    return fx*fx - z*z;
}

static float gf_torus(float x, float y, float z) {
    float r;
    xform(&x, &y, &z, mat_random);
    r = sqrtf(x*x + z*z) - 0.6f;
    return 0.0625f - r*r - y*y;
}

static float gf_plane(float x, float y, float z) {
    xform(&x, &y, &z, mat_random);
    return 0.1f - fabsf(z);
}

static float gf_ball(float x, float y, float z) {
    x += 0.5f * sinf(0.923f * t + 5.233f);
    y += 0.5f * sinf(0.427f * t + 3.622f);
    z += 0.5f * sinf(0.241f * t + 1.562f);
    return 0.375f - x*x - y*y - z*z;
}

#define LOGO_FUNC(group, xyscale, zscale) \
    static float gf_logo_##group(float x, float y, float z) { \
        xform(&x, &y, &z, mat_follow); \
        return getlogo(y * (xyscale), z * (xyscale), x * (zscale), logo_##group); \
    }

LOGO_FUNC(still, 10.0f, 10.0f)
LOGO_FUNC(drg,    8.0f,  8.0f)
LOGO_FUNC(fm,     8.0f,  8.0f)
LOGO_FUNC(mds,    9.0f,  8.0f)

const static grid_func grid_func_list[] = {
    gf_logo_drg,
    gf_ball,
    gf_logo_mds,
    gf_plane,
    gf_logo_still,
    gf_torus,
    gf_logo_fm,
NULL};
static const grid_func *current_grid_func = grid_func_list;
static float next_switch = SWITCH_INTERVAL;

#define GP_ENABLED  1
#define GP_VISIBLE  2
#define GP_FADING   4

typedef struct _grid_point {
    unsigned char ix, iy, iz, flags;
    float         cx, cy, cz, light;
} grid_point_t;

static grid_point_t grid[GRID_SIZE * GRID_SIZE * GRID_SIZE];
static float gc[GRID_SIZE];

static void InitGrid(void) {
    int x, y, z;
    grid_point_t *p = grid;
    for (z = 0;  z < GRID_SIZE;  ++z) {
        gc[z] = z * (2.0f / (GRID_SIZE - 1)) - 1.0f;
    for (y = 0;  y < GRID_SIZE;  ++y) {
    for (x = 0;  x < GRID_SIZE;  ++x) {
        p->flags = 0;
        p->ix = x;
        p->iy = y;
        p->iz = z;
    ++p; } } }
}

static void ClearGrid(void) {
    int x, y, z;
    grid_point_t *p = grid;
    for (z = 0;  z < GRID_SIZE;  ++z) {
    for (y = 0;  y < GRID_SIZE;  ++y) {
    for (x = 0;  x < GRID_SIZE;  ++x) {
        p->flags = 0;
        p->cx = (float) x;
        p->cy = (float) y;
        p->cz = (float) z;
        p->light = 0.0f;
    ++p; } } }
}

static void shuffle_list(int *list, int size) {
    int pos, temp;
    while (size > 1) {
        pos = randNext() % size;
        if (pos) {
            temp = list[pos];
            list[pos] = list[0];
            list[0] = temp;
        }
        --size;
        ++list;
    }
}

static void UpdateGrid(grid_func func, float dt) {
    grid_point_t *p, *q;
    int x, y, z, i;
    float dt2;
    static int list[GRID_SIZE * GRID_SIZE * GRID_SIZE];
    int *ss_list = list;
    int *sd_list = &list[GRID_SIZE * GRID_SIZE * GRID_SIZE - 1];
    int ss_count = 0, sd_count = 0;

    /***** apply grid function and detect changes *****/
    p = grid;  i = 0;
    for (z = 0;  z < GRID_SIZE;  ++z) {
    for (y = 0;  y < GRID_SIZE;  ++y) {
    for (x = 0;  x < GRID_SIZE;  ++x) {
        int fl = p->flags;
        int de = (func(gc[x], gc[y], gc[z]) > 0.0f) - (fl & GP_ENABLED);
        if (de > 0) {
            // point becomes enabled -> add to "searching source" list
            p->flags = fl | GP_ENABLED;
            ss_list[ss_count++] = i;
        }
        else if (de < 0) {
            // point becomes disabled -> add to "searching destination" list
            p->flags = fl & (~GP_ENABLED);
            *sd_list-- = i;
            ++sd_count;
        }
    ++p; ++i; } } }
//OUT2("ss_count=%d, sd_count=%d -> ", ss_count, sd_count);

    /***** shuffle lists *****/
    shuffle_list(ss_list, ss_count);
    shuffle_list(sd_list, sd_count);

    /***** partner search *****/
    while (ss_count && sd_count) {
        int best_partner = 0;
        p = &grid[*ss_list++];
        // TODO: search
        q = &grid[sd_list[best_partner]];
        // copy source to destination, activate destination, kill source
        p->cx = q->cx;
        p->cy = q->cy;
        p->cz = q->cz;
        p->light = q->light;
        p->flags = GP_ENABLED | GP_VISIBLE | GP_FADING;
        q->flags = 0;
        // reduce "searching destination" list
        sd_list[best_partner] = sd_list[0];
        ++sd_list;
        --sd_count;
    }
//OUT2("ss_count=%d, sd_count=%d\n", ss_count, sd_count);

    /***** enable orphaned invisible "searching source" points *****/
    while (ss_count--) {
        p = &grid[*ss_list++];
        if (!(p->flags & GP_VISIBLE)) {
            p->flags = GP_ENABLED | GP_VISIBLE | GP_FADING;
            p->cx = p->ix;
            p->cy = p->iy;
            p->cz = p->iz;
            p->light = 0.0f;
        }
    }

    /***** update positions and intensities *****/
    if (dt > 1.0f) dt = 1.0f;
    dt2 = 1.0f - dt;
    p = grid;
    for (i = GRID_SIZE * GRID_SIZE * GRID_SIZE;  i;  --i, ++p) {
        float l;
        if (!(p->flags & GP_VISIBLE)) continue;
        l = p->light * dt2;
        if (p->flags & GP_ENABLED) {
            l += dt;
            if (l >= (1.0f - 1.0f / 256))
                p->flags &= ~GP_FADING;
        } else {
            p->flags |= GP_FADING;
            if (l < (1.0f / 256)) {
                p->flags = 0;  // just became invisible
                continue;
            }
        }
        p->light = l;
        p->cx = p->cx * dt2 + p->ix * dt;
        p->cy = p->cy * dt2 + p->iy * dt;
        p->cz = p->cz * dt2 + p->iz * dt;
    }
}

static void prepare_cubes(void) {
    InitGrid();
    ClearGrid();
}

static float last_t = 0.0f;

static void run_cubes(void) {
    // sub-scene switch
    if (t > next_switch) {
        if (!*(++current_grid_func))
            current_grid_func = grid_func_list;
        next_switch += SWITCH_INTERVAL;
    }

    // grid update
    randSeed((unsigned long) (t * 184723.4828f));
    UpdateGrid(*current_grid_func, 1.0f - expf(MOVE_SPEED * (last_t - t)));
    last_t = t;
}

#undef GRID_SIZE


////////////////////////////////////////////////////////////////////////////////
// Vortex 2 Magnets Scene
////////////////////////////////////////////////////////////////////////////////

#define GRID_SIZE    32
#define GRID_SIZE_X  GRID_SIZE
#define GRID_SIZE_Y  (GRID_SIZE + (GRID_SIZE+2)/4 - (GRID_SIZE+8)/16)

#define PHYS_TIME_STEP    (1.0 / 600)
#define MAGNETISM         ((float)(10.0 * GRID_SIZE * PHYS_TIME_STEP))
#define MAGNET_SPEED      0.5f
#define MAGNET_ACCEL      0.0625f
#define MAGNET_RANGE      (0.4f * GRID_SIZE)
#define SPOT_SIZE         (GRID_SIZE / 8.0f)
#define SPOT_ATTENUATION  (1.0f * (float)PHYS_TIME_STEP)
#define DOT_SCALE         0.4f

#define CAMSWITCH_1_TIME   (POINT11-POINT10)
#define CAMSWITCH_1_DUR    4.0
#define CAMSWITCH_2_TIME   (POINT12-POINT10)
#define CAMSWITCH_2_DUR    2.0

static float FadeOut = 1.0f;

typedef struct {
    float pos[2];
    float dir[2];
    float spot;
    float light;
} magnet_grid_point_t;

#define GRID_POINT_COUNT (GRID_SIZE_X * GRID_SIZE_Y)
static magnet_grid_point_t points[GRID_POINT_COUNT];

static float norm2(float *v) {
    float n = isqrtf(v[0]*v[0] + v[1]*v[1]);
    v[0] *= n;
    v[1] *= n;
    return n;
}

static float t_phys;

static void GetDotPos(float *pos, float t_raw) {
    float t = (float)t_raw;
    t *= 1.0f + t * MAGNET_ACCEL;
    t *= MAGNET_SPEED;
    pos[0] = MAGNET_RANGE * (
             0.654f * sinf(0.842f * t + 2.345f)
           + 0.321f * sinf(1.233f * t + 6.789f)
           + 0.025f * sinf(4.732f * t + 4.321f));
    pos[1] = MAGNET_RANGE * (
             0.543f * sinf(1.427f * t + 0.123f)
           + 0.210f * sinf(0.923f * t + 4.567f)
           + 0.247f * sinf(2.184f * t + 8.901f));
}


static void UpdateMagnetGrid(float *dot) {
    magnet_grid_point_t *p = points;
    int i;
    float d[2], r, n;
    for (i = GRID_POINT_COUNT;  i;  --i, ++p) {
        // compute distance vector
        d[0] = dot[0] - p->pos[0];
        d[1] = dot[1] - p->pos[1];

        // compute distance, normalize distance vector
        r = d[0]*d[0] + d[1]*d[1];
        #define MIN_DIST (1.0f / 65536)
        if (r < MIN_DIST) r = MIN_DIST;

        // compute weight and update distance vector
        n = MAGNETISM / (r*r) * isqrtf(r);
        p->dir[0] += d[0] * n;
        p->dir[1] += d[1] * n;
        norm2(p->dir);

        // update lightness
        r = sqrtf(r) * (1.0f / SPOT_SIZE);
        r = sqrtf(r);
        if (r >= 1.0f) r = 1.0f;
        n = 1.0f - r;
        r *= 1.0f - SPOT_ATTENUATION;
        p->light = r * p->light + n * 2.0f;
    }
}

static void DrawGrid(void) {
    magnet_grid_point_t *p = points;
    int i;
    float r, g, b, l, dx, dy;
    for (i = GRID_POINT_COUNT;  i;  --i, ++p) {
        // set up color stuff
        l = p->light;
        r = 0.75f + 0.25f * p->dir[0];
        g = 0.75f + 0.25f * p->dir[1];
        if (l > 1.0f) {
            l -= 1.0f;
            b = 1.0f - l;
            r = b * r + l;
            g = b * g + l;
            b = 1.0f;
        } else
            b = 0.5f + 0.5f * l;
        l = p->spot * (float)FadeOut;
        fake3(r*l, g*l, b*l);

        // compute side vector and draw the thing
        dx = DOT_SCALE * (p->dir[0] + p->dir[1]);
        dy = DOT_SCALE * (p->dir[1] - p->dir[0]);
        fake2(p->pos[0] + dx, p->pos[1] + dy);
        fake2(p->pos[0] + dy, p->pos[1] - dx);
        fake2(p->pos[0] - dx, p->pos[1] - dy);
        fake2(p->pos[0] - dy, p->pos[1] + dx);
    }
}

static void prepare_magnet(void) {
    int x, y;
    magnet_grid_point_t *p = points;
    randSeed(823475);
    for (y = 0;  y < GRID_SIZE_Y;  ++y) {
        float py = (y - (0.5f * GRID_SIZE_Y)) * 0.866025403784439f;
        float px = (-0.5f * GRID_SIZE_X + 0.25f) + ((y & 1) * 0.5f);
        for (x = GRID_SIZE_X;  x;  --x, ++p) {
            float r = sqrtf(px*px + py*py);
            p->pos[0] = px;
            p->pos[1] = py;
            p->dir[0] = (float)randFloatAroundZero(1.0);
            p->dir[1] = (float)randFloatAroundZero(1.0);
            norm2(p->dir);
            r = 1.0f - (r / (0.5f * GRID_SIZE));
            if (r < 0.0f) r = 0.0f;
            if (r > 1.0f) r = 1.0f;
            p->spot = sqrtf(r);
            p->light = 0.0f;
            px += 1.0f;
        }
    }
}

static void run_magnet(void) {
    float curr[2];
    t_phys = 0.25f * t;
    GetDotPos(curr, t);
    UpdateMagnetGrid(curr);
    DrawGrid();
}


////////////////////////////////////////////////////////////////////////////////
// 8-Bit Wonderland Flowers
////////////////////////////////////////////////////////////////////////////////

#define FLOWER_COUNT      2048
#define FLOWER_XY_SPEED   5.0f
#define FLOWER_Z_SPEED    10.0f
#define FLOWER_A_SPEED    1.0f
#define FLOWER_GRAVITY    2.0f
#define FLOWER_SCALE      0.5f

typedef struct _flower {
    float x0, y0, z0;
    float vx, vy, vz;
    float a0[4];
    float va[4];
} flower_t;
static flower_t Flowers[FLOWER_COUNT];

static void CreateNewFlowers(float cx, float cy, float cz, float r) {
    flower_t *f = Flowers;
    int i, j;
    randSeed(63487649);
    for (i = FLOWER_COUNT;  i;  --i, ++f) {
        f->x0 = cx + (float)randFloatAroundZero(1.0) * r;
        f->y0 = cy + (float)randFloatAroundZero(1.0) * r;
        f->z0 = cz + (float)randFloatAroundZero(1.0) * r * 0.5f;
        f->vx = (float)randFloatAroundZero(FLOWER_XY_SPEED);
        f->vy = (float)randFloatAroundZero(FLOWER_XY_SPEED);
        f->vz = (float)randFloat() * (-FLOWER_Z_SPEED);
        for (j = 0;  j < 3;  ++j) {
            f->a0[j] = (float)randFloat() * 6.2f;
            f->va[j] = (float)randFloatAroundZero(FLOWER_A_SPEED);
        }
    }
}

static void DrawFlowers(float t, float *v) {
    flower_t *f = Flowers;
    int i;
    float a, b, c;
    float ax, ay, az;
    float bx, by, bz;
    float cx, cy, cz;
    for (i = FLOWER_COUNT;  i;  --i, ++f) {
        // compute center of flower
        cx = f->x0 + t *  f->vx;
        cy = f->y0 + t *  f->vy;
        cz = f->z0 + t * (f->vz + t * FLOWER_GRAVITY);

        // compute first "arm"
        a = f->a0[0] + t * f->va[0];
        b = f->a0[1] + t * f->va[1];
        c = cosf(a);
        ax = c * sinf(b);
        ay = c * cosf(b);
        az = sinf(a);

        // compute second "arm"
        a = f->a0[2] + t * f->va[2];
        b = f->a0[3] + t * f->va[3];
        c = cosf(a);
        bx = c * sinf(b);
        by = sinf(a);
        bz = c * cosf(b);

        // orthogonalize
        c = ax*bx + ay*by + az*bz;
        bx -= ax * c;
        by -= ay * c;
        bz -= az * c;

        // scale arms
        bx *= FLOWER_SCALE;
        by *= FLOWER_SCALE;
        bz *= FLOWER_SCALE;
        ax *= FLOWER_SCALE;
        ay *= FLOWER_SCALE;
        az *= FLOWER_SCALE;

        // create vertices
        v[ 0] = 0.0f;   v[ 1] = 0.0f;
        v[ 2] = cx+ax;  v[ 3] = cy+ay;  v[ 4] = cz+az;
        v[ 5] = 0.0f;   v[ 6] = 1.0f;
        v[ 7] = cx+bx;  v[ 8] = cy+by;  v[ 9] = cz+bz;
        v[10] = 1.0f;   v[11] = 1.0f;
        v[12] = cx-ax;  v[13] = cy-ay;  v[14] = cz-az;
        v[15] = 1.0f;   v[16] = 0.0f;
        v[17] = cx-bx;  v[18] = cy-by;  v[19] = cz-bz;
        v += 20;
    }
}

static void prepare_flowers(void) {
    CreateNewFlowers(0.0f, 0.0f, -11.25f, 5.0f);
}

static float vertexbuf[FLOWER_COUNT * 20];
static void run_flowers(void) {
    DrawFlowers(t, vertexbuf);
    vertexbuf[10] = vertexbuf[5] * vertexbuf[2];  // force ICL to not optimize vertexbuf away
}

////////////////////////////////////////////////////////////////////////////////
// Main Benchmark Code
////////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
    static double getTime(void) {
        return 0.001 * GetTickCount();
    }
#else
    static double getTime(void) {
        static struct timeval tv;
        gettimeofday(&tv, NULL);
        return (tv.tv_sec - 1000000000) + 0.000001 * tv.tv_usec;
    }
#endif

#define DO_BENCH(name, repeat, description) \
    do { \
        printf("%-40s", #repeat "x " description); \
        fflush(stdout); \
        prepare_##name(); \
        t = 0.0f; \
        tstart = getTime(); \
        for (i = repeat;  i;  --i) { \
            run_##name(); \
            t += 1.0f / 64; \
        } \
        printf("%20.1f fps\n", repeat / (getTime() - tstart)); \
    } while (0)


int main(int argc, char *argv[]) {
    double tstart;
    unsigned int i;
    DO_BENCH(cubes,   500,       "Vortex 2 Cube Grid");
    DO_BENCH(magnet,  20000,     "Vortex 2 Magnet Grid");
    DO_BENCH(flowers, 5000,      "8-Bit Wonderland Flowers");
    return 0;
}
